#  _________________________________________________________________________
#
#  Coopr: A COmmon Optimization Python Repository
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  For more information, see the Coopr README.txt file.
#  _________________________________________________________________________

__all__ = ['Suffix','active_export_suffix_generator','active_import_suffix_generator']

import sys
import logging
import weakref
import copy
import pyutilib.math
from six import iteritems, iterkeys, itervalues

from coopr.pyomo.base.component import Component, register_component

logger = logging.getLogger('coopr.pyomo')

# A list of convenient suffix generators, including:
#   - active_export_suffix_generator **(used by problem writers)
#   - export_suffix_generator
#   - active_import_suffix_generator **(used by OptSolver and PyomoModel._load_solution)
#   - import_suffix_generator
#   - active_local_suffix_generator
#   - local_suffix_generator
#   - active_suffix_generator
#   - suffix_generator

def active_export_suffix_generator(a_block,datatype=False):
    if (datatype is False):
        for name, suffix in iteritems(a_block.active_subcomponents(Suffix)):
            if suffix.exportEnabled() is True:
                yield name, suffix
    else:
        for name, suffix in iteritems(a_block.active_subcomponents(Suffix)):
            if (suffix.exportEnabled() is True) and \
               (suffix.getDatatype() is datatype):
                yield name, suffix

def export_suffix_generator(a_block,datatype=False):
    if (datatype is False):
        for name, suffix in iteritems(a_block.subcomponents(Suffix)):
            if suffix.exportEnabled() is True:
                yield name, suffix
    else:
        for name, suffix in iteritems(a_block.subcomponents(Suffix)):
            if (suffix.exportEnabled() is True) and \
               (suffix.getDatatype() is datatype):
                yield name, suffix

def active_import_suffix_generator(a_block,datatype=False):
    if (datatype is False):
        for name, suffix in iteritems(a_block.active_subcomponents(Suffix)):
            if suffix.importEnabled() is True:
                yield name, suffix
    else:
        for name, suffix in iteritems(a_block.active_subcomponents(Suffix)):
            if (suffix.importEnabled() is True) and \
               (suffix.getDatatype() is datatype):
                yield name, suffix

def import_suffix_generator(a_block,datatype=False):
    if (datatype is False):
        for name, suffix in iteritems(a_block.subcomponents(Suffix)):
            if suffix.importEnabled() is True:
                yield name, suffix
    else:
        for name, suffix in iteritems(a_block.subcomponents(Suffix)):
            if (suffix.importEnabled() is True) and \
               (suffix.getDatatype() is datatype):
                yield name, suffix

def active_local_suffix_generator(a_block,datatype=False):
    if (datatype is False):
        for name, suffix in iteritems(a_block.active_subcomponents(Suffix)):
            if suffix.getDirection() is Suffix.LOCAL:
                yield name, suffix
    else:
        for name, suffix in iteritems(a_block.active_subcomponents(Suffix)):
            if (suffix.getDirection() is Suffix.LOCAL) and \
               (suffix.getDatatype() is datatype):
                yield name, suffix

def local_suffix_generator(a_block,datatype=False):
    if (datatype is False):
        for name, suffix in iteritems(a_block.subcomponents(Suffix)):
            if suffix.getDirection() is Suffix.LOCAL:
                yield name, suffix
    else:
        for name, suffix in iteritems(a_block.subcomponents(Suffix)):
            if (suffix.getDirection() is Suffix.LOCAL) and \
               (suffix.getDatatype() is datatype):
                yield name, suffix

def active_suffix_generator(a_block,datatype=False):
    if (datatype is False):
        for name, suffix in iteritems(a_block.active_subcomponents(Suffix)):
            yield name, suffix
    else:
        for name, suffix in iteritems(a_block.active_subcomponents(Suffix)):
            if suffix.getDatatype() is datatype:
                yield name, suffix

def suffix_generator(a_block,datatype=False):
    if (datatype is False):
        for name, suffix in iteritems(a_block.subcomponents(Suffix)):
            yield name, suffix
    else:
        for name, suffix in iteritems(a_block.subcomponents(Suffix)):
            if suffix.getDatatype() is datatype:
                yield name, suffix

# THOUGHTS: 

# TODO: Should we, when setting the value on a component, start collecting
#       a dictionary of components by their type (like on Blocks)

# TODO: Do we enforce that a component must be at least declared
#       on a suffix before the default suffix value is associated
#       with it? Another possibility is to explicitly add the component
#       to the internal values dictionary when first accessed, given
#       that the suffix has a default value.

# TODO: Should calling clearValue on a parent component (like _VarArray) clear
#       its own dictionary entry AS WELL AS any possible dictionary entries
#       corresponding to its subcomponent entries (e.g. _VarData members)?
#       Right now this does not happen. One could ask the same of setValue.

# TODO: Can we get rid of the datatype specifier? Or at least only segregate between
#       Symbolic (require a table) vs Numeric (no table) suffixes. Distinguishing 
#       between int, float, and bool seems to be very unpythonic.

class Suffix(Component):
    """A model suffix, representing extranious model data"""

    """
    Constructor Arguments:
        name        The name of this suffix
        direction   The direction of information flow for this suffix.
                        By default, this is LOCAL, indicating that no
                        suffix data is exported or imported.
        datatype    A variable type associated with all values of this
                        suffix. By default, this is float. **Note that
                        all values assigned to this suffix will be 
                        casted using the function associated with
                        this keyword, unless the suffix datatype is
                        assigned None.
        default     The default value associated with all components
                        declared on this suffix.
        table       A dictionary like object mapping datatype values
                        to numeric (integer) values. The associated integer
                        values are used when exporting/importing suffix 
                        data. **Note: this keyword is required for symbolic (str)
                        datatype suffixes.
        rule        A 
    """

    # If more directions are added be sure to update the error
    # message in the setDirection method
    LOCAL  = 0 # neither
    EXPORT = 1 # sent to solver or other external location
    IMPORT = 2 # obtained from solver or other external source
    IMPORT_EXPORT = 3 # both
    SuffixDirections = (LOCAL,EXPORT,IMPORT,IMPORT_EXPORT)
    SuffixDatatypes = (bool,int,float,str,None)

    def __getstate__(self):
        """
        This method must be defined for deepcopy/pickling because 
        this class relies on component ids.
        """
        result = super(Suffix, self).__getstate__()
        result['_component_value_map'] = tuple((ref(),val) for ref,val in itervalues(self._component_value_map) if ref() is not None)
        return result

    def __setstate__(self, state):
        """
        This method must be defined for deepcopy/pickling because 
        this class relies on component ids.
        """
        state['_component_value_map'] = dict((id(comp),(weakref.ref(comp),val)) for comp,val in state['_component_value_map'])
        return super(Suffix,self).__setstate__(state)

    def __init__(self, *args, **kwds):
        
        # Suffix type information
        self._rule = None
        self._direction = None
        self._datatype = None
        self._reverse_table = None
        self._table = None
        self._default = None
        # The meat of suffixes... a dictionary mapping 
        # Pyomo component ids to suffix values
        self._component_value_map = {}

        # TODO: __FIX_ME__: Blocks assume everything is a (Sparse)IndexedComponent inside
        #       __setattr__ so until this is fixed we need to define a dummy _index data
        #       member
        self._index = None

        # The suffix direction
        direction = kwds.pop('direction',Suffix.LOCAL)
        # The suffix datatype
        datatype = kwds.pop('datatype',float)
        # The optional suffix table
        table = kwds.pop('table',None)
        # The default suffix value for all components
        default = kwds.pop('default',None)
        # A constructor rule for convenience with AbstractModels
        self._rule = kwds.pop('rule',None)
        
        # Initialize base class
        kwds.setdefault('ctype', Suffix)
        Component.__init__(self, *args, **kwds)

        # Check that keyword values make sense 
        # (these function have internal error checking).
        self.setDirection(direction)
        self.setDatatype(datatype,table)
        self.setDefault(default)

        if self._rule is None:
            self.construct()

    def reset(self):
        """
        Reconstructs this component by clearing all values and re-calling
        construction rule if it exists
        """
        self.clearValue()
        self._constructed = False
        self.construct()

    def construct(self, data=None):
        """
        Constructs this component, applying rule if it exists.
        """
        generate_debug_messages = (__debug__ is True) and (logger.isEnabledFor(logging.DEBUG) is True)

        if generate_debug_messages is True:
            logger.debug("Constructing suffix %s",self.cname())

        if self._constructed is True:
            return
        self._constructed = True
        if self._rule is not None:
            self._rule(self._parent())
            return

    def exportEnabled(self):
        """
        Returns True when this suffix is enabled for export to solvers.
        """
        return bool(self._direction & Suffix.EXPORT)

    def importEnabled(self):
        """
        Returns True when this suffix is enabled for import from solutions.
        """
        return bool(self._direction & Suffix.IMPORT)

    def getRule(self):
        """
        Returns the current constructor rule on this component."
        """
        return self._rule

    def setRule(self,rule):
        """
        Sets the constructor rule for this component.
        """
        self._rule = rule
        
    def setNumericValue(self,component,value):
        """
        Set the current numeric value of this suffix on the
        specified component. This method should be used by all
        solution loaders that hold a numeric interpretation of 
        the suffix.
        """
        if self._reverse_table is not None:
            if value.__class__ is not int:
                raise ValueError("Can not call Suffix.setNumericValue() with a non-integer type "\
                                  "when the Suffix component has a table. Current value %s has "\
                                  "type %s." % (value, value.__class__))
            self.setValue(component,self._reverse_table[value])
            return

        # this should not be possible at this point
        # str datatype should always have a reverse table
        assert self._datatype is not str

        self.setValue(component,value)

    def getNumericValue(self,component):
        """
        Returns the current numeric value of this suffix on the 
        specified component. This method should be used by all
        problem writers that require a numeric interpretation of
        the suffix. If no value exists and no default value exists
        this function will return None.
        """
        if (self._datatype is None) and (self._table is None):
            raise RuntimeError("Suffix method getNumericValue can not be called on a suffix "\
                                "component whose datatype and table are both None. Only a suffix "\
                                "table can guarantee numeric return type.")

        value = self.getValue(component)
        if value is None:
            return None

        if self._table is not None:
            return self._table[value]

        # this should not be possible at this point
        # str datatype should always have a table
        assert self._datatype is not str

        if self._datatype is bool:
            return int(self.getValue(component))

        return self.getValue(component)

    def updateValues(self,data_buffer):
        """
        Updates the suffix data given a list of component/value tuples. Provides
        an improvement in efficiency over calling setValue on every component.
        """
        if self._datatype is None:
            self._component_value_map.update((id(comp),(weakref.ref(comp),val)) for comp,val in data_buffer)
        else:
            self._component_value_map.update((id(comp),(weakref.ref(comp),self._datatype(val))) for comp,val in data_buffer)

    def setValue(self,*args):
        """
        Sets the value of this suffix on the specified component 
        or on all components if no component is specified. This 
        function accepts exactly one or two arguments. A suffix
        value is always a required argument. If a component
        is specified, it must be the first argument.
        """
        if len(args) == 1:
            # setting the value on all components currently handled
            value = args[0]
            # This helps clear up any ambiguity about what it means for a suffix
            # to be defined for a component. Otherwise we would have to ask questions like:
            #  (1) Is it directly referenced in the internal suffix dict?
            #  (2) Is it indirectly referenced (Vararray -> VarData) in the internal suffix dict?
            #  (3) If default is not None, does it mean this suffix is defined for all components?
            # I can't answer these right now, so it's easiest to say: Obtaining a value of None
            # by calling getValue implies the suffix is NOT defined for a component.
            if value is None:
                raise ValueError("Explicitly setting component suffix values "\
                                  "to None is not allowed. Try calling the suffix method "\
                                  "clearValue() instead, and ensure that the suffix "\
                                  "default value is set to None.")
            if self._datatype is None:
                self._component_value_map = dict((id(ref()),(ref,value)) for ref,val in itervalues(self._component_value_map) \
                                                                               if ref() is not None)
            else:
                self._component_value_map = dict((id(ref()),(ref,self._datatype(value))) for ref,val in itervalues(self._component_value_map) \
                                                                                               if ref() is not None)
        elif len(args) == 2:
            # setting the value on a specified component
            component = args[0]
            value = args[1]
            # This helps clear up any ambiguity about what it means for a suffix
            # to be defined for a component. Otherwise we would have to ask questions like:
            #  (1) Is it directly referenced in the internal suffix dict?
            #  (2) Is it indirectly referenced (Vararray -> VarData) in the internal suffix dict?
            #  (3) If default is not None, does it mean this suffix is defined for all components?
            # I can't answer these right now, so its easiest to say: Obtaining a value of None
            # by calling getValue implies the suffix is NOT defined for a component.
            if value is None:
                raise ValueError("Explicitly setting component suffix values "\
                                  "to None is not allowed. Try calling the suffix method "\
                                  "clearValue() instead, and ensure that the suffix "\
                                  "default value is set to None.")
            cid = id(component)
            if self._datatype is None:
                self._component_value_map[cid] = (weakref.ref(component),value)
            else:
                self._component_value_map[cid] = (weakref.ref(component),self._datatype(value))
        else:
            raise TypeError("Suffix.setValue() takes exactly one or two arguments (%s given)" % (len(args),))

    def getValue(self,component):
        """
        Returns the current value of this suffix on the 
        specified component.
        """
        cid = id(component)
        # check if the component has been assigned an individual
        # suffix value.
        # **Note: We do this safely by checking that the id is still valid
        #         for the original component it was assigned to by checking
        #         the weakref
        value = None
        if cid in self._component_value_map:
            value = self._component_value_map[cid]
            if value[0]() is component:
                value = value[1]
            else:
                # A rare case where the id was reassigned
                value = None
                del self._component_value_map[cid]
        if (value is None) and (not (component is component.component())):     
            # In the event that this component is part of an indexed
            # component and its container component exists (e.g.
            # x[1] has not been assigned a suffix value but x has), then
            # let us return that suffix value. This mocks behavior of
            # allowing for a default suffix value to exist across an
            # indexed component.
            component = component.component()
            cid = id(component)
            if cid in self._component_value_map:
                value = self._component_value_map[cid]
                if value[0]() is component:
                    value = value[1]
                else:
                    # A rare case where the id was reassigned
                    value = None
                    del self._component_value_map[cid]
        if value is None:
            # Then lets go ahead and return the full suffix default,
            # which could be None
            return self._default
        return value
        
    def clearValue(self,*args):
        """
        Clears suffix information for a single component, or all
        components if no component is specified. This function 
        accepts, at most, one argument, which is assumed to be
        the component. 
        """
        if len(args) == 0:
            # clearing the value on all components currently handled
            self._component_value_map = {}
        elif len(args) == 1:
            # clearing the value on a specified component
            cid = id(args[0])
            if cid in self._component_value_map:
                value = self._component_value_map[cid]
                if value[0]() is args[0]:
                    del self._component_value_map[cid]
        else:
            raise TypeError("Suffix.clearValue() takes exactly one or zero arguments (%s given)" % (len(args),))

    def setTable(self,table):
        """
        Set the suffix table. Symbolic type suffixes (str)
        require that the table keyword defines a dictionary
        mapping string values to integers. Setting table
        to None, clears the existing table if possible.
        """
        if (self._datatype is str) and (table is None):
            raise ValueError("Suffix table is required when suffix datatype is of type %s." % (str,))
        if table is not None:
            reverse_table = {}
            for key,value in iteritems(table):
                if (self._datatype is not None) and (not key.__class__ is self._datatype):
                    raise TypeError("Suffix table must have keys of type %s, but key %s has type %s" \
                        % (self._datatype, key, key.__class__))
                if not value.__class__ is int:
                    raise TypeError("Suffix table must have values of type %s, but value %s has type %s" \
                    % (int,value, value.__class__))
                reverse_table[value] = key
            if len(table) != len(reverse_table):
                raise ValueError("Error constructing reverse suffix table. Suffix table values must all" \
                                  "be unique.")
            self._table = copy.deepcopy(table)
            self._reverse_table = reverse_table
        else:
            self._table = None
            self._reverse_table = None

    def getTable(self):
        """
        Return a copy of the suffix table map."
        """
        return copy.deepcopy(self._table)

    def getReverseTable(self):
        """
        Return a copy of the suffix table reverse map."
        """
        return copy.deepcopy(self._reverse_table)
        
    def setDatatype(self,datatype,table=None):
        """
        Set the suffix datatype. Symbolic type suffixes (str)
        require that the table keyword defines a dictionary
        mapping string values to integers. Setting table
        to None or ommiting the table keyword clears the 
        existing table if possible.
        """
        if datatype not in self.SuffixDatatypes:
            raise ValueError("Suffix datatype must be one of: %s. \n" \
                              "Value given: %s" % (Suffix.SuffixDatatypes,datatype))
        self._datatype = datatype
        self.setTable(table)

    def getDatatype(self):
        """
        Return the suffix datatype.
        """
        return self._datatype

    def setDirection(self,direction):
        """ 
        Set the suffix direction.
        """
        if not direction in self.SuffixDirections:
            raise ValueError("Suffix direction must be one of: %s. \n" \
                              "Value given: %s" % (('Suffix.EXPORT','Suffix.IMPORT','Suffix.IMPORT_EXPORT','Suffix.LOCAL'),direction))
        self._direction = direction

    def getDirection(self):
        """
        Return the suffix direction.
        """
        return self._direction

    def setDefault(self,default):
        """
        Set the default value for this suffix on all
        components.
        """
        if default is None:
            self._default = None
        else:
            if self._datatype is None:
                self._default = default
            else:
                self._default = self._datatype(default)
            
    def getDefault(self):
        """
        Return the suffix default.
        """
        return self._default

    def __str__(self):
        """
        Return a string representation of the suffix.  If the name attribute is None,
        then return ''
        """
        if self.cname() is None:
            return ""
        else:
            return self.cname()

    def pprint(self, ostream=None, verbose=False):
        """Print component information"""
        if ostream is None:
            ostream = sys.stdout
        ostream.write("  "+self.cname()+" :")
        if not self.doc is None:
            ostream.write(self.doc+'\n')
        else:
            ostream.write('\n')
        ostream.write("  ")
        ostream.write("\tDirection=")
        if self._direction == Suffix.EXPORT:
            ostream.write("EXPORT\n")
        elif self._direction is Suffix.IMPORT:
            ostream.write("IMPORT\n")
        elif self._direction is Suffix.LOCAL:
            ostream.write("LOCAL\n")
        elif self._direction is Suffix.IMPORT_EXPORT:
            ostream.write("IMPORT_EXPORT\n")
        else:
            raise ValueError("Unexpected Suffix direction encountered in pprint method.")
        ostream.write("\tDatatype= "+str(self._datatype)+'\n')
        ostream.write("\tTable= "+str(self._table)+'\n')
        ostream.write("\tDefault= "+str(self._default)+'\n')
        # print all component values
        if verbose is True:
            ostream.write('\tValues=\n')
            name_data = dict((ref().cname(True),val) for ref,val \
                                                     in itervalues(self._component_value_map) \
                                                     if ref() is not None)
            for key in sorted(iterkeys(name_data)):
                ostream.write('\t  %s = %s\n' % (key, name_data[key]))

register_component(Suffix, "Extraneous model data")



#  _________________________________________________________________________
#
#  Coopr: A COmmon Optimization Python Repository
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  For more information, see the Coopr README.txt file.
#  _________________________________________________________________________

__all__ = ['Component', 'cname']

import weakref
import sys
from six import iteritems
from coopr.pyomo.base.plugin import register_component


def _cname_index_generator(idx):
    if idx.__class__ is tuple:
        return "[" + ",".join(str(i) for i in idx) + "]"
    else:
        return "[" + str(idx) + "]"


def cname(component, index=None, fully_qualified=False):
    base = component.cname(fully_qualified=fully_qualified)
    if index is None:
        return base
    else:
        if index not in component.index_set():
            raise KeyError( "Index %s is not valid for component %s"
                            % (index, component.cname(True)) )
        return base + _cname_index_generator( index )


class Component(object):
    """
    This is the base class for all Pyomo modeling components.

    Constructor arguments:
        ctype           The class type for the derived subclass
        doc             A text string describing this component
        name            A name for this component

    Public class attributes:
        active          A boolean that is true if this component will be 
                            used to construct a model instance
        doc             A text string describing this component

    Private class attributes:
        _constructed    A boolean that is true if this component has been
                            constructed
        _parent         A weakref to the parent block that owns this component
        _type           The class type for the derived subclass
    """

    def __init__ (self, **kwds):
        #
        # Get arguments
        #
        self.doc   = kwds.pop('doc', None)
        self.name  = kwds.pop('name', str(type(self).__name__)) # "{unnamed}"
        self._type = kwds.pop('ctype', None)
        if kwds:
            raise ValueError(
                "Unexpected keyword options found while constructing '%s':\n\t%s"
                % ( type(self).__name__, ','.join(sorted(kwds.keys())) ))
        #
        # Verify that ctype has been specified.
        #
        if self._type is None:
            raise DeveloperError("Must specify a class for the component type!")
        #
        self.active = True
        self._constructed = False
        self._parent = None    # Must be a weakref

    def __getstate__(self):
        """
        This method must be defined to support pickling because this class
        owns weakrefs for '_parent'.
        """
        # Nominally, __getstate__() should return:
        #
        # state = super(Class, self).__getstate__()
        # for i in Class.__dict__:
        #     state[i] = getattr(self,i)
        # return state
        #
        # Hoewever, in this case, the (nominal) parent class is
        # 'object', and object does not implement __getstate__.  Since
        # super() doesn't actually return a class, we are going to check
        # the *derived class*'s MRO and see if this is the second to
        # last class (the last is always 'object').  If it is, then we
        # can allocate the state dictionary.  If it is not, then we call
        # the super-class's __getstate__ (since that class is NOT
        # 'object').
        if self.__class__.__mro__[-2] is Component:
            state = dict(self.__dict__)
        else:
            state = super(Component,self).__getstate__()
            for key,val in iteritems(self.__dict__):
                if key not in state:
                    state[key] = val
        
        if self._parent is not None:
            state['_parent'] = self._parent()
        return state

    def __setstate__(self, state):
        """
        This method must be defined to support pickling because this class
        owns weakrefs for '_parent'.
        """
        if state['_parent'] is not None and \
                type(state['_parent']) is not weakref.ref:
            state['_parent'] = weakref.ref(state['_parent'])
        # Note: our model for setstate is for derived classes to modify
        # the state dictionary as control passes up the inheritance
        # hierarchy (using super() calls).  All assignment of state ->
        # object attributes is handled at the last class before 'object'
        # (which may -- or may not (thanks to MRO) -- be here.
        if self.__class__.__mro__[-2] is Component:
            for key, val in iteritems(state):
                # Note: per the Python data model docs, we explicitly
                # set the attribute using object.__setattr__() instead
                # of setting self.__dict__[key] = val.
                object.__setattr__(self, key, val)
        else:
            return super(Component,self).__setstate__(state)

    def __str__(self):
        return self.cname()

    def activate(self):
        """Set the active attribute to True"""
        self.active=True

    def deactivate(self):
        """Set the active attribute to False"""
        self.active=False

    def type(self):
        """Return the class type for this component"""
        return self._type

    def construct(self, data=None):                     #pragma:nocover
        """API definition for constructing components"""
        pass

    def is_constructed(self):                           #pragma:nocover
        """Return True if this class has been constructed"""
        return self._constructed

    def valid_model_component(self):
        """Return True if this can be used as a model component."""
        return True

    def pprint(self, ostream=None, verbose=False):
        """Print component information"""
        if ostream is None:
            ostream = sys.stdout
        ostream.write("  %s:"%(self.name))

    def component(self):
        return self

    def parent(self):
        if self._parent is None:
            return None
        else:
            return self._parent()

    def model(self):
        ans = self.parent()
        if ans is None:
            return None
        while ans.parent() is not None:
            ans = ans.parent()
        return ans

    def cname(self, fully_qualified=False, name_buffer=None):
        if fully_qualified and self.parent() != self.model():
            return self.parent().cname(fully_qualified, name_buffer) \
                + "." + self.name
        return self.name

    # a simple utility to return an id->index dictionary for
    # all composite ComponentData instances.
    def id_index_map(self):
        result = dict()
        for index, component_data in iteritems(self):
            result[id(component_data)] = index
        return result

class ComponentData(object):

    __slots__ = ( '_component', )

    def __init__(self, owner):
        if owner is None:
            self._component = None
        else:
            self._component = weakref.ref(owner)

    def __getstate__(self):
        # Nominally, __getstate__() should return:
        #
        # state = super(Class, self).__getstate__()
        # for i in Class.__slots__:
        #    state[i] = getattr(self,i)
        # return state
        #
        # Hoewever, in this case, the (nominal) parent class is
        # 'object', and object does not implement __getstate__.  Since
        # super() doesn't actually return a class, we are going to check
        # the *derived class*'s MRO and see if this is the second to
        # last class (the last is always 'object').  If it is, then we
        # can allocate the state dictionary.  If it is not, then we call
        # the super-class's __getstate__ (since that class is NOT
        # 'object').
        #
        # Further, since there is only a single slot, and that slot
        # (_component) requires special processing, we will just deal
        # with it explicitly.  As _component is a weakref (not
        # pickable), so we need to resolve it to a concrete object.
        if self.__class__.__mro__[-2] is ComponentData:
            state = {}
        else:
            state = super(ComponentData,self).__getstate__()
        
        if self._component is None:
            state['_component'] = None
        else:
            state['_component'] = self._component()
        return state

    def __setstate__(self, state):
        # FIXME: We shouldn't have to check for weakref.ref here, but if
        # we don't the model cloning appears to fail (in the Benders
        # example)
        if state['_component'] is not None and \
                type(state['_component']) is not weakref.ref:
            state['_component'] = weakref.ref(state['_component'])

        # Note: our model for setstate is for derived classes to modify
        # the state dictionary as control passes up the inheritance
        # hierarchy (using super() calls).  All assignment of state ->
        # object attributes is handled at the last class before 'object'
        # (which may -- or may not (thanks to MRO) -- be here.
        if self.__class__.__mro__[-2] is ComponentData:
            for key, val in iteritems(state):
                # Note: per the Python data model docs, we explicitly
                # set the attribute using object.__setattr__() instead
                # of setting self.__dict__[key] = val.
                object.__setattr__(self, key, val)
        else:
            return super(ComponentData,self).__setstate__(state)

    def __str__(self):
        return self.cname()

    def component(self):
        if self._component is None: 
            return None
        return self._component()

    # returns the index of this ComponentData instance relative
    # to the parent component index set. None is returned if 
    # this instance does not have a parent component, or if
    # - for some unknown reason - this instance does not belong
    # to the parent component's index set. not intended to be
    # a fast method - should be used rarely, primarily in 
    # cases of label formulation.
    def index(self):
        self_component = self.component()
        if self_component is None:
            return None
        for idx, component_data in self_component.iteritems():
            if id(component_data) == id(self):
                return idx
        return None
        
    def parent(self):
        ans = self.component()
        if ans is None:
            return None
        # Directly call the Component's model() to prevent infinite
        # recursion for singleton objects.
        if ans is self:
            return super(ComponentData, ans).parent()
        else:
            return ans.parent()

    def cname(self, fully_qualified=False, name_buffer=None):
        c = self.component()
        if c is self:
            return super(ComponentData, self).cname(fully_qualified, name_buffer)

        base = c.cname(fully_qualified, name_buffer)
        if name_buffer is not None:
            if id(self) in name_buffer:
                # should we delete the entry to save mamory?  JDS:
                # probably not, because for nested blocks, we would hit
                # this part of the code numerous times.
                return name_buffer[id(self)]
            for idx, obj in iteritems(c._data):
                name_buffer[id(obj)] = base + _cname_index_generator(idx)
            if id(self) in name_buffer:
                return name_buffer[id(self)]
        else:
            for idx, obj in iteritems(c._data):
                if obj is self:
                    return base + _cname_index_generator(idx)

        raise RuntimeError("Fatal error: cannot find the component data in "
                           "the owning component's _data dictionary.")

    def model(self):
        ans = self.component()
        if ans is None:
            return None
        # Directly call the Component's model() to prevent infinite
        # recursion for singleton objects.
        if ans is self:
            return super(ComponentData, ans).model()
        else:
            return ans.model()


class DeveloperError(Exception):
    """
    Exception class used to throw errors that result from
    programming errors, rather than user modeling errors (e.g., a
    component not declaring a 'ctype').
    """

    def __init__(self, val):
        self.parameter = val

    def __str__(self):                                  #pragma:nocover
        return repr(self.parameter)




from coopr.pyomo.check.hooks.base import *
from coopr.pyomo.check.hooks.model import *
from coopr.pyomo.check.hooks.function import *

__all__ = ['IPreCheckHook', 'IPostCheckHook']
__all__.extend(['ModelTrackerHook'])
__all__.extend(['FunctionTrackerHook'])



from coopr.pyomo.check.checkers.pyomo.model import *
from coopr.pyomo.check.checkers.pyomo.imports import *
from coopr.pyomo.check.checkers.pyomo.rule import *
from coopr.pyomo.check.checkers.pyomo.conditional import *
from coopr.pyomo.check.checkers.pyomo.assignment import *



#
# Test the Pyomo command-line interface
#

import unittest
import os
import sys
from os.path import abspath, dirname, join
currdir = dirname(abspath(__file__))+os.sep
scriptdir = dirname(dirname(dirname(dirname(dirname(dirname(abspath(__file__)))))))+os.sep
scriptdir = join(scriptdir,'examples','pyomo','piecewise')

import coopr.pyomo
import coopr.pyomo.scripting.convert as convert
import pyutilib.th as unittest

_NL_diff_tol = 1e-9
_LP_diff_tol = 1e-9

class Test(unittest.TestCase):

    def run_convert2nl(self, name):
        os.chdir(currdir)
        return convert.pyomo2nl(['--symbolic-solver-labels','--skip-canonical-repn',join(scriptdir,name)])

    def run_convert2lp(self, name):
        os.chdir(currdir)
        return convert.pyomo2lp(['--symbolic-solver-labels',join(scriptdir,name)])

    def test_step_lp(self):
        """Test examples/pyomo/piecewise/step.py"""
        self.run_convert2lp('step.py')
        self.assertFileEqualsBaseline(join(currdir,'unknown.lp'), currdir+'step.lp', tolerance=_LP_diff_tol)

    def test_step_nl(self):
        """Test examples/pyomo/piecewise/step.py"""
        self.run_convert2nl('step.py')
        self.assertFileEqualsBaseline(join(currdir,'unknown.nl'), currdir+'step.nl', tolerance=_NL_diff_tol)
        os.remove(join(currdir,'unknown.row'))
        os.remove(join(currdir,'unknown.col'))

    def test_affine_lp(self):
        """Test examples/pyomo/piecewise/affine.py"""
        self.run_convert2lp('affine.py')
        self.assertFileEqualsBaseline(join(currdir,'unknown.lp'), currdir+'affine.lp', tolerance=_LP_diff_tol)

    def test_affine_nl(self):
        """Test examples/pyomo/piecewise/affine.py"""
        self.run_convert2nl('affine.py')
        self.assertFileEqualsBaseline(join(currdir,'unknown.nl'), currdir+'affine.nl', tolerance=_NL_diff_tol)
        os.remove(join(currdir,'unknown.row'))
        os.remove(join(currdir,'unknown.col'))

    def test_convex_lp(self):
        """Test examples/pyomo/piecewise/convex.py"""
        self.run_convert2lp('convex.py')
        self.assertFileEqualsBaseline(join(currdir,'unknown.lp'), currdir+'convex.lp', tolerance=_LP_diff_tol)

    def test_convex_nl(self):
        """Test examples/pyomo/piecewise/convex.py"""
        self.run_convert2nl('convex.py')
        self.assertFileEqualsBaseline(join(currdir,'unknown.nl'), currdir+'convex.nl', tolerance=_NL_diff_tol)
        os.remove(join(currdir,'unknown.row'))
        os.remove(join(currdir,'unknown.col'))

    def test_indexed_lp(self):
        """Test examples/pyomo/piecewise/indexed.py"""
        self.run_convert2lp('indexed.py')
        self.assertFileEqualsBaseline(join(currdir,'unknown.lp'), currdir+'indexed.lp', tolerance=_LP_diff_tol)

    def test_indexed_nl(self):
        """Test examples/pyomo/piecewise/indexed.py"""
        self.run_convert2nl('indexed.py')
        self.assertFileEqualsBaseline(join(currdir,'unknown.nl'), currdir+'indexed.nl', tolerance=_NL_diff_tol)
        os.remove(join(currdir,'unknown.row'))
        os.remove(join(currdir,'unknown.col'))

    def test_indexed_nonlinear_nl(self):
        """Test examples/pyomo/piecewise/indexed_nonlinear.py"""
        self.run_convert2nl('indexed_nonlinear.py')
        self.assertFileEqualsBaseline(join(currdir,'unknown.nl'), currdir+'indexed_nonlinear.nl', tolerance=_NL_diff_tol)
        os.remove(join(currdir,'unknown.row'))
        os.remove(join(currdir,'unknown.col'))


if __name__ == "__main__":
    unittest.main()



#  _________________________________________________________________________
#
#  Coopr: A COmmon Optimization Python Repository
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  For more information, see the Coopr README.txt file.
#  _________________________________________________________________________

"""
Block Properties:

1. Blocks are indexed components that contain other components
   (including blocks)

2. Blocks have a global attribute that defines whether construction is
   deferred.  This applies to all components that they contain except
   blocks.  Blocks contained by other blocks use their local attribute
   to determine whether construction is deferred.

NOTE: Blocks do not currently maintain statistics about the sets,
parameters, constraints, etc that they contain, including these
components from subblocks.
"""

__all__ = ['Block', 'active_subcomponents_generator', 'active_subcomponents_data_generator']

import sys
import weakref
import logging
from operator import itemgetter
from six import iterkeys, iteritems, itervalues, StringIO

from pyutilib.component.core import alias, ExtensionPoint
from pyutilib.misc import Container

from coopr.pyomo.base.plugin import *
from coopr.pyomo.base.component import Component, ComponentData
from coopr.pyomo.base.sets import Set, _SetContainer
from coopr.pyomo.base.rangeset import RangeSet
from coopr.pyomo.base.var import Var
from coopr.pyomo.base.constraint import Objective, Constraint
from coopr.pyomo.base.param import Param
from coopr.pyomo.base.misc import apply_indexed_rule
from coopr.pyomo.base.set_types import IntegerSet, BooleanSet
from coopr.pyomo.base.sparse_indexed_component import SparseIndexedComponent

logger = logging.getLogger('coopr.pyomo')

#
# We now include a number of generators for iterating over various
# components within a block.  the primary purpose is to streamline code,
# as this kind of iteration is done *everywhere* in Coopr. any generally
# useful iteration schemes should be moved into the following segment.
# 

# iterates over all active components (e.g., Var) of the specified type
# in the input block.  sorting, if specified, is by component name.
def active_subcomponents_generator(block, component_type, deterministic=False):
    compList = iteritems( block.active_subcomponents(component_type) )
    for name, component in compList:
        if component.active:
            yield component

# iterates over all active sub_components (e.g., _VarData) of all active
# components of the specified type in the input block.  If deterministic
# is set to True, the objects will be returned in a deterministic order
# (components in declaration order and within each component sorted by index)
def active_subcomponents_data_generator(block, component_type, deterministic=False):
    active_components = \
        active_subcomponents_generator( block, component_type, deterministic )
    for component in active_components:
        subCompList = iteritems(component._data)
        if deterministic:
            subCompList = sorted(subCompList, key=itemgetter(0))
        for name, componentData in subCompList:
            if componentData.active:
                yield componentData

# 
#
#

class _BlockConstruction(object):
    """
    This class holds a "global" dict used when constructing
    (hierarchical) models.
    """
    data = {}


class _BlockData(ComponentData):
    """
    This class holds the fundamental block data.
    """

    class PseudoMap(object):
        """
        This class presents a "mock" dict interface to the internal
        BlockComponents data structures.  We return this object to the
        user to preserve the historical "{ctype : {name : obj}}"
        interface without actually regenerating that dict-of-dicts data
        structure.
        """

        __slots__ = ( '_block', '_ctype', '_active' )
        
        def __init__(self, block, ctype, active=None):
            self._block = block
            self._ctype = ctype
            self._active = active

        def __iter__(self):
            return self.iterkeys()

        def __getitem__(self, key):
            if key in self._block._decl:
                x = self._block._decl_order[self._block._decl[key]]
                if self._ctype is None or x[0].type() is self._ctype:
                    if self._active is None or x[0].active == self._active:
                        return x[0]
            msg = ""
            if self._active is not None:
                msg += self._active and "active " or "inactive "
            if self._ctype is not None:
                msg += self._ctype.__name__ + " "
            raise KeyError( "%scomponent %s not found in block %s" 
                            % (msg, key, self._block.cname()))

        def __nonzero__(self):
            # Shortcut: this will bail after finding the first item
            for x in itervalues(self):
                return True
            return False

        __bool__ = __nonzero__

        def __len__(self):
            if self._active is None:
                if self._ctype is None:
                    return sum(x[2] for x in itervalues(self._block._ctypes))
                else:
                    return self._block._ctypes.get(self._ctype,(0,0,0))[2]
            #For active==True/False, we have to count by brute force
            ans = 0
            for x in itervalues(self):
                ans += 1
            return ans

        def __contains__(self, key):
            if key in self._block._decl:
                x = self._block._decl_order[self._block._decl[key]]
                if x[0].type() is self._ctype:
                    if self._active is None or x[0].active == self._active:
                        return True
            return False

        def iterkeys(self):
            if self._ctype is None:
                idx = 0
            else:
                idx = self._block._ctypes.get(self._ctype, [None])[0]
            while idx is not None:
                try:
                    x = self._block._decl_order[idx]
                except IndexError:
                    break

                if x[0] is not None:
                    if self._active is None or x[0].active == self._active:
                        yield x[0].name
                if self._ctype is None:
                    idx += 1
                else:
                    idx = x[1]

        def itervalues(self):
            if self._ctype is None:
                idx = 0
            else:
                idx = self._block._ctypes.get(self._ctype, [None])[0]
            while idx is not None:
                try:
                    x = self._block._decl_order[idx]
                except IndexError:
                    break

                if x[0] is not None:
                    if self._active is None or x[0].active == self._active:
                        yield x[0]
                if self._ctype is None:
                    idx += 1
                else:
                    idx = x[1]

        def iteritems(self):
            if self._ctype is None:
                idx = 0
            else:
                idx = self._block._ctypes.get(self._ctype, [None])[0]
            while idx is not None:
                try:
                    x = self._block._decl_order[idx]
                except IndexError:
                    break

                if x[0] is not None:
                    if self._active is None or x[0].active == self._active:
                        yield (x[0].name, x[0])
                if self._ctype is None:
                    idx += 1
                else:
                    idx = x[1]

        def keys(self):
            return list(self.iterkeys())

        def values(self):
            return list(self.itervalues())

        def items(self):
            return list(self.iteritems())

    def __init__(self, owner):
        #
        # We used to define an internal structure that looked like:
        #
        #    _component    = { ctype : OrderedDict( name : obj ) }
        #    _declarations = OrderedDict( name : obj )
        #
        # This structure is convenient, but the overhead of carrying
        # around roughly 20 dictionaries for every block consumed a
        # nontrivial amount of memory.  Plus, the generation and
        # maintenance of OrderedDicts appeared to be disturbingly slow.
        #
        # We now "mock up" this data structure using 2 dicts and a list:
        #
        #    _ctypes     = { ctype : [ first idx, last idx, count ] }
        #    _decl       = { name  : idx }
        #    _decl_order = [ (obj, next_ctype_idx) ]
        #
        # Some notes: As Pyomo models rarely *delete* objects, we
        # currently never remove items from the _decl_order list.  If
        # the component is ever removed / cleared, we simply mark the
        # object as None.  If models crop up where we start seeing a
        # significant amount of adding / removing components, then we
        # can revisit this decision (although we will probably continue
        # marking entries as None and just periodically rebuild the list
        # as opposed to maintining the list without any holes.
        #
        ComponentData.__init__(self, owner)
        # Note: call super() here to bypass the Block __setattr__
        #   _ctypes:      { ctype -> [1st idx, last idx, count] }
        #   _decl:        { name -> idx }
        #   _decl_order:  list( tuples( obj, next_type_idx ) )
        super(_BlockData, self).__setattr__('_ctypes', {})
        super(_BlockData, self).__setattr__('_decl', {})
        super(_BlockData, self).__setattr__('_decl_order', [])


    def __getstate__(self):
        # Note: _BlockData is NOT slot-ized, so we must pickle the
        # entire __dict__.  However, we want the base class's
        # __getstate__ to override our blanket approach here (i.e., it
        # will handle the _component weakref).
        ans =  dict(self.__dict__)
        ans.update(ComponentData.__getstate__(self))
        return ans

    def __setstate__(self, state):
        # We want the base class's __setstate__ to override our blanket
        # approach here (i.e., it will handle the _component weakref).
        for (slot_name, value) in iteritems(state):
            super(_BlockData, self).__setattr__(slot_name, value)
        ComponentData.__setstate__(self, state)


    #def __getitem__(self, name):
    #    # This supports two key functionalities:
    #    #  1. block["component_name"] as a convenient alternative to
    #    #     block.component_name
    #    #  2. block[Var] as an alternative to block.subcomponents(Var)
    #
    #    if name in self._decl:
    #        return self._decl_order[self._decl[name]][0]
    #    elif name in self._ctypes:
    #        return BlockComponents.PseudoMap(self, name)
    #    elif isinstance(name,basestring):
    #        return None
    #    else:
    #        return {}

    def __setattr__(self, name, val):
        #
        # Set an attribute on this Block.  In general, the most common
        # case for this is setting a *new* attribute.  After that, there
        # is updating an existing Component value, with the least common
        # case being resetting an existing general attribute.  Given
        # that, we will:
        #
        # 1) handle new attributes, with special handling of Components
        if name not in self.__dict__:
            if isinstance(val, Component):
                self.add_component(name, val)
            else:
                super(_BlockData, self).__setattr__(name, val)
                #self.__dict__[name] = val
        # 2) Update the value of existing [singleton] Components
        # (through the component's set_value())
        elif name in self._decl:
            # JDS: setting a component to None should attempt
            #if val is None:
            #    self.del_component(name)
            if isinstance(val, Component):
                logger.warn(
                    "Attempting to implictly replace the Component attribute "
                    "%s (type=%s) on block %s with a new Component (type=%s)."
                    "\nThis is usually indicative of a modelling error.\n"
                    "To avoid this warning, use block.del_component() and "
                    "block.add_component()."
                    % ( name, type(self.subcomponent(name)), self.cname(True), 
                        type(val) ) )
                self.del_component(name)
                self.add_component(name, val)
            else:
                try:
                   self.subcomponent(name).set_value(val)
                except AttributeError:
                    logger.error(
                        "Expected component %s (type=%s) on block %s to have a "
                        "'set_value' method, but none was found." % 
                        ( name, type(self.subcomponent(name)), 
                          self.cname(True) ) )
                    raise
        # 3) handle setting non-Component attributes
        else:
            # NB: This is important: the _BlockData is either a
            # singleton Block (where _parent and _component are defined)
            # or a single block within an Indexed Block (where only
            # _component is defined).  Regardless, the __init__()
            # declares these attributes and sets them to either None or
            # a weakref.  This means that we will never have a problem
            # converting these objects from weakrefs into Blocks and
            # back (when pickling): because the attribute is already
            # in __dict__, we will not hit the add_component /
            # del_component branches above.  It also means that any
            # error checking we want to do when assigning these
            # attributes should be done here.

            # NB: isintance() can be slow; however, it is only slow when
            # it returns False.  Since the common paths on this branch
            # should return True, this shouldn't be too inefficient.
            if name == '_parent':
                if val is not None and not isinstance(val(), _BlockData):
                    raise ValueError(
                        "Cannot set the '_parent' attribute of Block '%s' "
                        "to a non-Block object (with type=%s); Did you "
                        "try to create a model component named '_parent'?" 
                        % (self.cname(), type(val)) )
                super(_BlockData, self).__setattr__(name, val)
                #self.__dict__[name] = val
            elif name == '_component':
                if val is not None and not isinstance(val(), _BlockData):
                    raise ValueError(
                        "Cannot set the '_component' attribute of Block '%s' "
                        "to a non-Block object (with type=%s); Did you "
                        "try to create a model component named '_component'?" 
                        % (self.cname(), type(val)) )
                super(_BlockData, self).__setattr__(name, val)
                #self.__dict__[name] = val
            # At this point, we should only be seeing non-component data
            # the user is hanging on the blocks (uncommon) or the
            # initial setup of the object data (in __init__).
            elif isinstance(val, Component):
                logger.warn(
                    "Attempting to reassign the non-component attribute %s "
                    "on block %s with a new Component with type %s.  This is "
                    "usually indicative of a modelling error."
                    % (name, self.cname(True), type(val)) )
                self.del_component(name)
                self.add_component(name, val)
            else:
                super(_BlockData, self).__setattr__(name, val)
                #self.__dict__[name] = val

 
    def __delattr__(self, name):
        # We need to make sure that del_component() gets called if a
        # user attempts to remove an attribute from a block that happens
        # to be a component
        if name in self._decl:
            self.del_component(name)
        else:
            super(_BlockData, self).__delattr__(name)


    def _add_temporary_set(self,val):
        """ TODO """
        # NB: only "old" IndexedComponent contain _index_set.
        #    SparseIndexedComponent has _implicit_subsets
        # FIXME: We need to make IndexedComponent look more like
        #    SparseIndexedComponent (or make it go away entirely)
        _component_sets = getattr(val, '_index_set', None)
        if _component_sets is None:
            _construct_index_set = False
            _component_sets = getattr(val, '_implicit_subsets', None)
        else:
            _construct_index_set = True

        # FIXME: The name attribute should begin with "_", and None
        # should replace "_unknown_"
        if _component_sets is not None:
            ctr=0
            for tset in _component_sets:
                if tset.cname() == "_unknown_":
                    self._construct_temporary_set(
                      tset,
                      val.cname()+"_index_"+str(ctr)
                    )
                ctr += 1
            if _construct_index_set:
                val._index = self._construct_temporary_set(
                    _component_sets,
                    val.cname()+"_index" )
        if isinstance(val._index,_SetContainer) and val._index.name == "_unknown_":
            self._construct_temporary_set(val._index,val.cname()+"_index")
        if getattr(val,'domain',None) is not None and val.domain.name == "_unknown_":
            self._construct_temporary_set(val.domain,val.cname()+"_domain")

    def _construct_temporary_set(self, obj, name):
        """ TODO """
        if type(obj) is tuple:
            if len(obj) == 1:                #pragma:nocover
                raise Exception(
                    "Unexpected temporary set construction for set "
                    "%s on block %s" % ( name, self.cname()) )
            else:
                tobj = obj[0]
                for t in obj[1:]:
                    tobj = tobj*t
                setattr(self,name,tobj)
                tobj.virtual=True
                return tobj
        if isinstance(obj,Set):
            setattr(self,name,obj)
            return obj


    def add_component(self, name, val):
        if not val.valid_model_component():
            raise RuntimeError(
                "Cannot add '%s' as a component to a model" % str(type(val)) )
        if name in self.__dict__:
            raise RuntimeError(
                "Cannot add component '%s' (type %s) to block '%s': a "
                "component by that name (type %s) is already defined."
                % (name, type(val), self.cname(), type(getattr(self, name))))

        # all Pyomo components have names. 
        val.name = name
        # all Pyomo components have Parents
        if val._parent is not None:
            logger.warn("Reassigning component %s to block %s"
                        % (val.cname(True), self.cname(True)) )
            val.parent().del_component(name)
        val._parent = weakref.ref(self)

        # We want to add the temporary / implicit sets first so that
        # they get constructed before this component
        self._add_temporary_set(val)

        # Add the component to the underlying Component store
        _new_idx = len(self._decl_order)
        self._decl[name] = _new_idx
        self._decl_order.append( (val, None) )

        # Add the component as an attribute
        super(_BlockData, self).__setattr__(name, val)
        #self.__dict__[name]=val

        # Update the ctype linked lists
        _type = val.type()
        if _type in self._ctypes:
            idx_info = self._ctypes[_type]
            tmp = idx_info[1]
            self._decl_order[tmp] = (self._decl_order[tmp][0], _new_idx)
            idx_info[1] = _new_idx
            idx_info[2] += 1
        else:
            self._ctypes[_type] = [_new_idx, _new_idx, 1]

        # Support implicit rule names
        if '_rule' in val.__dict__:
            if val._rule is None:
                frame = sys._getframe(2)
                locals_ = frame.f_locals
                if val.cname()+'_rule' in locals_:
                    val._rule = locals_[val.cname()+'_rule']
        # FIXME: This is a HACK to support the way old Blocks and legacy
        # IndexedComponents (like Set) behave.  In particular, Set does
        # not define a "rule" attribute.  I put the hack back in to get
        # some tests passing again, but in all honesty, I am amazed this
        # ever worked properly. [JDS]
        elif getattr(val, 'rule', None) is None:
            frame = sys._getframe(2)
            locals_ = frame.f_locals
            if val.cname()+'_rule' in locals_:
                val.rule = locals_[val.cname()+'_rule']

        # If the block is Concrete, construct the component
        if self.component()._constructed:
            # NB: we don't have to construct the temporary / implicit
            # sets here: if necessary, that happens when
            # _add_temporary_set() calls add_component().
            if id(self) in _BlockConstruction.data:
                data = _BlockConstruction.data[id(self)].get(name,None)
            else:
                data = None
            # This is tricky: If we are in the middle of constructing an
            # indexed block, the block component already has
            # _constructed=True.  Now, if the _BlockData.__init__()
            # defines any local variables (like coopr.gdp.Disjunct's
            # indicator_var), cname(True) will fail: this block data
            # exists and has a parent(), but it has not yet been added
            # to the parent's _data (so the idx lookup will fail in cname()).
            if __debug__ and logger.isEnabledFor(logging.DEBUG):
                if self.parent() is None:
                    _blockName = "[Model]"
                else:
                    try:
                        _blockName = "Block '%s'" % self.cname(True)
                    except:
                        _blockName = "Block '%s[...]'" \
                            % self.component().cname(True)
                logger.debug( "Constructing %s '%s' on %s from data=%s",
                              val.__class__.__name__, val.cname(), 
                              _blockName, str(data) )
            try:
                val.construct(data)
            except:
                err = sys.exc_info()[1]
                logger.error(
                    "Constructing component '%s' from data=%s failed:\n%s: %s",
                    str(val.cname(True)), str(data).strip(),
                    type(err).__name__, err )
                raise
            if __debug__ and logger.isEnabledFor(logging.DEBUG):
                if _blockName[-1] == "'":
                    _blockName = _blockName[:-1] + '.' + val.cname() + "'"
                else:
                    _blockName = "'" + _blockName + '.' + val.cname() + "'"
                _out = StringIO()
                val.pprint(ostream=_out)
                logger.debug( "Constructed component '%s':\n%s" 
                              % ( _blockName, _out.getvalue() ) )


    def del_component(self, name):
        # FIXME: Is this necessary?  Should this raise an exception?
        if name not in self._decl:
            return

        # Remove the attribute
        del self.__dict__[name]

        # Replace the component in the master list
        idx = self._decl[name]
        del self._decl[name]
        obj = self._decl_order[idx][0]
        self._decl_order[idx] = (None, self._decl_order[idx][1])

        # Update the ctype linked lists
        ctype_info = self._ctypes[obj.type()]
        ctype_info[2] -= 1
        if ctype_info[2] == 0:
            del self._ctypes[obj.type()]

        # Clear the _parent attribute
        obj._parent = None

    def reclassify_subcomponent_type(self, name_or_object, new_ctype):
        try:
            c = name_or_object.component()
        except AttributeError:
            c = self.subcomponent(name_or_object)
        if c._type is not new_ctype:
            self.del_component(c.cname())
            c._type = new_ctype
            self.add_component(c.cname(), c)


    #def keys(self):
    #    return list(self.iterkeys())

    #def __iter__(self):
    #    for x in self._decl_order:
    #        if x[0] is not None:
    #            yield x[0].cname()

    #def iterkeys(self):
    #    for x in self._decl_order:
    #        if x[0] is not None:
    #            yield x[0].cname()

    #def itervalues(self):
    #    for x in self._decl_order:
    #        if x[0] is not None:
    #            yield x[0]

    #def iteritems(self):
    #    for x in self._decl_order:
    #        if x[0] is not None:
    #            yield (x[0].cname(), x[0])

    def contains_component(self, ctype):
        return ctype in self._ctypes and self._ctypes[ctype][2]

    def subcomponent_map(self, ctype=None, active=None):
        """
        Return information about the block components.  If ctype is
        None, return the dictionary that maps
           {component type -> {name -> instance}}
        Otherwise, return the dictionary that maps
           {name -> instance}
        for the specified component type.

        Note: the actual {name->instance} object is a PseudoMap that
        implements a lightweight interface to the underlying
        BlockComponents data structures.
        """

        if ctype is None:
            ans = {}
            for x in self._ctypes:
                ans[x] = _BlockData.PseudoMap(self, x, active)
            return ans
        else:
            return _BlockData.PseudoMap(self, ctype, active)

    def components(self, ctype=None, active=None):
        #logger.warn("DEPRECATION WARNING: Block.components() has been "
        #            "renamed Block.subcomponents().")
        return self.subcomponents(ctype, active)

    def active_components(self, ctype=None):
        #logger.warn("DEPRECATION WARNING: Block.active_components() has been "
        #            "renamed Block.active_subcomponents().")
        return self.active_subcomponents(ctype)

    def subcomponent(self, name):
        if name in self._decl:
            return self._decl_order[self._decl[name]][0]
        else:
            return None

    def subcomponents(self, ctype=None, active=None):
        return _BlockData.PseudoMap(self, ctype, active)
            
    def active_subcomponents(self, ctype=None):
        """
        Returns the active components in this block.  Return values
        match those of .subcomponents().
        """
        return self.subcomponents(ctype, True)

    def fix_all_vars(self):
        for var in itervalues(self.subcomponents(Var)):
            var.fix_all()
        for block in itervalues(self.subcomponents(Block)):
            block.fix_all_vars()

    def unfix_all_vars(self):
        for var in itervalues(self.subcomponents(Var)):
            var.unfix_all()
        for block in itervalues(self.subcomponents(Block)):
            block.unfix_all_vars()

    def is_constructed(self):
        """
        A boolean indicating whether or not all *active* components of the
        input model have been properly constructed.
        """
        if not self._constructed:
            return False
        for x in self._decl_order:
            if x[0] is not None and x[0].active and not x[0].is_constructed():
                return False
        return True

    def pprint(self, filename=None, ostream=None, verbose=False):
        """
        Print a summary of the model info
        """
        if ostream is None:
            ostream = sys.stdout
        if filename is not None:
            OUTPUT=open(filename,"w")
            self.pprint(ostream=OUTPUT, verbose=verbose)
            OUTPUT.close()
            return
        if ostream is None:
            ostream = sys.stdout
        #
        # We hard-code the order of the core Pyomo modeling
        # components, to ensure that the output follows the logical order
        # that expected by a user.
        #
        items = [Set, RangeSet, Param, Var, Objective, Constraint, Block]
        for item in ExtensionPoint(IModelComponent):
            if not item in items:
                items.append(item)

        # Currently, model components are not actually being registered
        # with the IModelComponent extension point (1 Nov 2010), so as a
        # workaround, we will loop through the components and add any
        # new components that actually have defined members.
        extra_items = []
        for ctype, info in iteritems(self._ctypes):
            if ctype not in items and info[2]:
                extra_items.append(ctype)
        # extra items get added alphabetically (so output is consistent)
        if sys.version_info < (3,0):
            items.extend(sorted(extra_items, key=lambda x: x.__name__))
        else:
            items.extend(sorted(extra_items, key=lambda x: x.__name__))

        for item in items:
            keys = list(iterkeys(self.subcomponents(item)))
            keys.sort()

            #
            # NOTE: these conditional checks should not be hard-coded.
            #
            ostream.write(str(len(keys))+' '+item.__name__+" Declarations\n")
            for key in keys:
                self.subcomponent(key).pprint(ostream=ostream, verbose=verbose)
            ostream.write("\n")
        #
        # Model Order
        #
        decl_order_keys = list(iterkeys(self.subcomponents()))
        ostream.write(str(len(decl_order_keys))+" Declarations:")
        for x in decl_order_keys:
            ostream.write(' '+str(x))
        ostream.write("\n")


class Block(SparseIndexedComponent):
    """
    A block in an optimization model.  By default, this defers construction of
    components until data is loaded.
    """

    def __new__(cls, *args, **kwds):
        if cls != Block:
            return super(Block, cls).__new__(cls)
        if args == ():
            return SingletonBlock.__new__(SingletonBlock)
        else:
            return IndexedBlock.__new__(IndexedBlock)


    def __init__(self, *args, **kwargs):
        """Constructor"""
        self._rule = kwargs.pop('rule', None )
        self._options = kwargs.pop('options', None )
        kwargs.setdefault('ctype', Block)
        SparseIndexedComponent.__init__(self, *args, **kwargs)

    def _default(self, idx):
        return self._data.setdefault(idx, _BlockData(self))

    def concrete_mode(self):
        """Configure block to immediately construct components"""
        self.construct()

    def symbolic_mode(self):
        """Configure block to defer construction of components"""
        if self._constructed:
            logger.error(
                "Attempting to return a Concrete model to Symbolic (Abstract) "
                "mode: you will likely experience unexpected and potentially "
                "erroneous behavior.")
        self._constructed=False

    def all_blocks(self, active=None, sort_by_keys=False):
        if active is not None and self.active != active:
            return
        blocks = iteritems(self)
        if sort_by_keys:
            blocks = sorted(blocks, key=itemgetter(0))

        for idx, b in blocks:
            yield b
            for subblock in itervalues(b.subcomponents(Block, active)):
                for tmp in subblock.all_blocks(active, sort_by_keys):
                    yield tmp

    # flags *all* active variables and their composite _VarData objects
    # as stale.  fixed variables are flagged as non-stale. the state of
    # all inactive variables is left unchanged.
    def flag_vars_as_stale(self):
        for block in self.all_blocks():
            for variable in active_subcomponents_generator(block,Var):
                variable.flag_as_stale()


    def find_component(self, label):
        """ TODO """
        cList = label.split('.')
        obj = self
        while cList:
            c = cList.pop(0).split(':')
            if len(c) > 1:
                idx = c[1].split(',')
                for i, val in enumerate(idx):
                    if val[0]=='#':
                        idx[i] = int(val[1:])
                    elif val[0]=='$':
                        idx[i] = val[1:]
                    elif val[0]=='!':
                        idx[i] = None
                    else:
                        # last-ditch effort to find things
                        try:
                            tmp = int(val)
                            idx[i] = tmp
                        except ValueError:
                            pass
                        #raise ValueError(
                        #    "missing index type specifier for index %s in canonical label %s" % (val,label) )
                if len(idx) > 1:
                    idx = tuple(idx)
                else:
                    idx = idx[0]
            else:
                idx = None
            try:
                if idx is None:
                    obj = getattr(obj, c[0])
                else:
                    obj = getattr(obj, c[0])[idx]
            except AttributeError:
                return None
        return obj


    def construct(self, data=None):
        """ TODO """
        if __debug__ and logger.isEnabledFor(logging.DEBUG):
            logger.debug( "Constructing %s '%s', from data=%s",
                          self.__class__.__name__, self.cname(), str(data) )
        if self._constructed:
            return
        self._constructed = True

        # We must check that any pre-existing components are
        # constructed.  This catches the case where someone is building
        # a Concrete model by building (potentially pseudo-abstract)
        # sub-blocks and then adding them to a Concrete model block.
        for idx in self._data:
            _block = self[idx]
            for name, obj in iteritems( _block.subcomponents() ):
                if not obj._constructed:
                    if data is None:
                        _data = None
                    else:
                        _data = data.get(name, None)
                    obj.construct(_data)
        
        if self._rule is None:
            return
        # If we have a rule, fire the rule for all indices.  
        # Notes:
        #  - Since this block is now concrete, any components added to
        #    it will be immediately constructed by
        #    block.add_component().
        #  - Since the rule does not pass any "data" on, we build a
        #    singleton "stack" of pointers to block data
        #    (_BlockConstruction.data) that the individual blocks'
        #    add_component() can refer back to to handle subcomponent
        #    construction.
        for idx in self._index:
            _block = self[idx]
            if data is not None and idx in data:
                _BlockConstruction.data[id(_block)] = data[idx]
            obj = apply_indexed_rule(
                None, self._rule, _block, idx, self._options )
            if id(_block) in _BlockConstruction.data:
                del _BlockConstruction.data[id(_block)]

            # TBD: Should we allow skipping Blocks???
            #if obj is Block.Skip and idx is not None:
            #   del self._data[idx]

    def display(self, filename=None, ostream=None):
        """
        Print the Pyomo model in a verbose format.
        """
        if filename is not None:
            OUTPUT=open(filename,"w")
            self.display(ostream=OUTPUT)
            OUTPUT.close()
            return
        if ostream is None:
            ostream = sys.stdout
        if self._parent is not None and self._parent() is not None:
            ostream.write("Block "+self.cname()+'\n')
        else:
            ostream.write("Model "+self.cname()+'\n')
        #
        ostream.write("\n")
        ostream.write("  Variables:\n")
        VAR = self.active_subcomponents(Var)
        if not VAR:
            ostream.write("    None\n")
        else:
            for obj in itervalues(VAR):
                obj.display(prefix="    ",ostream=ostream)
        #
        ostream.write("\n")
        ostream.write("  Objectives:\n")
        OBJ = self.active_subcomponents(Objective)
        if not OBJ:
            ostream.write("    None\n")
        else:
            for obj in itervalues(OBJ):
                obj.display(prefix="    ",ostream=ostream)
        ostream.write("\n")
        #
        CON = self.active_subcomponents(Constraint)
        ostream.write("  Constraints:\n")
        if not CON:
            ostream.write("    None\n")
        else:
            for obj in itervalues(CON):
                obj.display(prefix="    ",ostream=ostream)


class SingletonBlock(_BlockData, Block):

    def __init__(self, *args, **kwds):
        _BlockData.__init__(self, self)
        Block.__init__(self, *args, **kwds)

    def __getstate__(self):
        ans = Block.__getstate__(self)
        ans2 = _BlockData.__getstate__(self)
        for attr, val in iteritems(ans2):
            if attr not in ans:
                ans[attr] = val
            elif val is not ans[attr]:
                if type(val) is weakref.ref:
                    pass
                elif type(ans[attr]) is weakref.ref:
                    ans[attr] = val
                else:
                    raise RuntimeError(
                        "Unexpected attribute difference when pickling "
                        "block attribute '%s' for block '%s'" 
                        % (attr, self.cname()) )
        return ans

    def __setstate__(self, state):
        # Important: call Block before _BlockData so that the Block
        # hierarchy has an opportunity to correct weakrefs before
        # _BlockData performs a blanket assignment
        Block.__setstate__(self, state)
        _BlockData.__setstate__(self, state)


class IndexedBlock(Block):

    def __init__(self, *args, **kwds):
        Block.__init__(self, *args, **kwds)

    #def display(self, ostream=None):
    #    self.pprint(ostream=ostream)

    #def __str__(self):
    #    return str(self.cname())


#register_component(Block, "Blocks are indexed components that contain "
#                   "one or more other model components.")



#  _________________________________________________________________________
#
#  Coopr: A COmmon Optimization Python Repository
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  _________________________________________________________________________


# Transformation heirarchy
from coopr.pyomo.transform.transformation import *
from coopr.pyomo.transform.linear_transformation import *
from coopr.pyomo.transform.nonlinear_transformation import *
from coopr.pyomo.transform.abstract_transformation import *
from coopr.pyomo.transform.concrete_transformation import *
from coopr.pyomo.transform.isomorphic_transformation import *
from coopr.pyomo.transform.nonisomorphic_transformation import *

# Transformations
import coopr.pyomo.transform.relax_integrality
import coopr.pyomo.transform.eliminate_fixed_vars
from coopr.pyomo.transform.standard_form import *
from coopr.pyomo.transform.equality_transform import *
from coopr.pyomo.transform.nonnegative_transform import *
from coopr.pyomo.transform.dual_transformation import DualTransformation
import coopr.pyomo.transform.util



from six import itervalues
from coopr.pyomo import *
from coopr.opt import SolverFactory
import yaml

opt = SolverFactory('cplexamp',solve_io='nl')

kwds = {'pw_constr_type':'UB','pw_repn':'DCC','sense':maximize,'force_pw':True}

problem_names = []
problem_names.append("piecewise_multi_vararray")
problem_names.append("concave_multi_vararray1")
problem_names.append("concave_multi_vararray2")
problem_names.append("convex_multi_vararray1")
problem_names.append("convex_multi_vararray2")
problem_names.append("convex_vararray")
problem_names.append("concave_vararray")
problem_names.append("convex_var")
problem_names.append("concave_var")
problem_names.append("piecewise_var")
problem_names.append("piecewise_vararray")
problem_names.append("step_var")
problem_names.append("step_vararray")

problem_names = ['convex_var']

for problem_name in problem_names:
    p = __import__(problem_name)
    
    model = p.define_model(**kwds)
    inst = model.create()

    results = opt.solve(inst,tee=True)

    inst.load(results)
    
    res = dict()
    for block in inst.all_blocks():
        for variable in itervalues(block.active_subcomponents(Var)):
            for var in itervalues(variable):
                name = var.cname(True)
                if (name[:2] == 'Fx') or (name[:1] == 'x'):
                    res[name] = value(var)
    print(res)

    #with open(problem_name+'_baseline_results.yml','w') as f:
    #    yaml.dump(res,f)
